# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-08-25 18:41
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0008_auto_20170509_1736'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transfer',
            name='event',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='transfers', to='pinax_stripe.Event'),
        ),
    ]
